/*
 * HFS (aka MAC) filesystem/handler for MorphOS/AmigaOS
 * Author: Marek 'March' Szyprowski <marek@amiga.pl>
 */

#include <exec/types.h>
#include <exec/execbase.h>
#include <exec/errors.h>

#include <dos/dos.h>
#include <dos/filehandler.h>
#include <devices/trackdisk.h>

#include <proto/exec.h>
#include <proto/dos.h>

#include "mac_fs.h"
#include "mac_protos.h"


LONG InitFSSM(struct FileSysStartupMsg *fssm_src)
{
	struct FileSysStartupMsg *fssm;
	struct DosEnvec *de, *de_src = BADDR(fssm_src->fssm_Environ);

	if ((fssm = AllocVecPooled(glob->mempool, sizeof(*fssm))) != NULL)
	{
		if ((de	= AllocVecPooled(glob->mempool, sizeof(*de))) != NULL)
		{
			memcpy(fssm, fssm_src, sizeof(*fssm));
			fssm->fssm_Environ = MKBADDR(de);
			memcpy(de, de_src, sizeof(*de));
			glob->fssm = fssm;
			/* fssm_Device and de_Control are not copied! */
			return 0;
		}
		FreeVecPooled(glob->mempool, fssm);
	}
	return ERROR_NO_FREE_STORE;
}

void FreeFSSM()
{
	struct DosEnvec *de = BADDR(glob->fssm->fssm_Environ);
	FreeVecPooled(glob->mempool, de);
	FreeVecPooled(glob->mempool, glob->fssm);
}

LONG UpdateDiskGeometry(void)
{
	struct DosEnvec *de = BADDR(glob->fssm->fssm_Environ);
	ULONG sector_size, sector_count;
	LONG err;

	kprintf("\n\tDisk size autodetect enabled, updating geometry\n");

	if ((err = GetDiskGeometry(&sector_size, &sector_count)) != 0)
		return err;

	kprintf("\tRead drive geometry - sector size: %ld, total sector count: %ld\n", sector_size, sector_count);

	/* update geometry */
	de->de_LowCyl = 0;
	de->de_SizeBlock = sector_size / 4;
	de->de_Surfaces = 1;
	de->de_BlocksPerTrack = 1;
	de->de_HighCyl = sector_count-1;

	SendDiskChangePacket(ACTION_DOSENVEC_UPDATE, NULL);
	return 0;
}

LONG GetDiskProtectionStatus(void)
{
   LONG res = ID_WRITE_PROTECTED;

	glob->diskioreq->iotd_Req.io_Command = TD_PROTSTATUS;
	glob->diskioreq->iotd_Req.io_Length = 0;
	glob->diskioreq->iotd_Req.io_Flags = IOF_QUICK;
	glob->diskioreq->iotd_Req.io_Actual = 0;
	glob->diskioreq->iotd_Req.io_Data = NULL;

	if (DoIO((struct IORequest*)glob->diskioreq) == 0)
	{
		if (glob->diskioreq->iotd_Req.io_Actual == 0)
		{
			/* Disk is not write protected */
			res = ID_VALIDATED;
		}
	}
	return res;
}

LONG GetDiskGeometry(ULONG *res_sector_size, ULONG *res_total_sectors)
{
	struct DosEnvec *de = BADDR(glob->fssm->fssm_Environ);
	LONG err;
	ULONG sector_size, total_sectors;
	struct DriveGeometry geo;

	glob->diskioreq->iotd_Req.io_Command = TD_GETGEOMETRY;
	glob->diskioreq->iotd_Req.io_Length = sizeof(struct DriveGeometry);
	glob->diskioreq->iotd_Req.io_Flags = 0;
	glob->diskioreq->iotd_Req.io_Actual = 0;
	glob->diskioreq->iotd_Req.io_Data=  &geo;
	geo.dg_SectorSize =  0;

	if ((err = DoIO((struct IORequest*)glob->diskioreq)) == 0)
	{
		switch(geo.dg_SectorSize)
		{
			case 512:
			case 1024:
			case 2048:
			case 4096:
			case 8192:
			case 16384:
			case 32768:
			{
				sector_size = geo.dg_SectorSize;
				total_sectors = geo.dg_TotalSectors;
                kprintf("\tDisk geometry - blocksize: %ld\n", geo.dg_SectorSize);
				break;
			}
			case 0:
			{
				kprintf("\tInvalid block size 0 reported - reading block size from mountlist: %ld\n", de->de_SizeBlock * 4);
				sector_size = de->de_SizeBlock * 4;
				total_sectors = geo.dg_TotalSectors;
				break;
			}
			default:
			{
				kprintf("\tInvalid blocksize found: %ld\n", geo.dg_SectorSize);
				return ERROR_OBJECT_WRONG_TYPE;
			}
		}
	}
	else
	{
		kprintf("\tReading drive geometry failed - reading block size from mountlist: %ld\n", de->de_SizeBlock * 4);
		sector_size = de->de_SizeBlock * 4;
		total_sectors = de->de_Surfaces * de->de_BlocksPerTrack * de->de_HighCyl;
	}
	if (res_sector_size)
		*res_sector_size = sector_size;
	if (res_total_sectors)
		*res_total_sectors = total_sectors;

	return 0;
}

struct IntData
{
#ifdef __MORPHOS__
	struct EmulLibEntry InterruptFunc;
#endif
	struct Interrupt Interrupt;
	struct Globals *glob;
	struct Task *task;
	ULONG signal;
};

#ifdef __MORPHOS__
static BOOL DiskChangeIntHandler(void)
{
	struct IntData *MyIntData=(struct IntData*) REG_A1;
#else
#define reg(x) __asm(#x)
static LONG DiskChangeIntHandler(struct IntData *MyIntData reg(a1))
{
#endif
//	  struct Globals *glob=MyIntData->glob;

	Signal(MyIntData->task, MyIntData->signal);
	return 0;
}

LONG InitDiskHandler (struct FileSysStartupMsg *fssm, ULONG *diskchgsig_bit)
{
	LONG err;
	ULONG diskchgintbit, flags, unit;
	UBYTE *device;

	unit = fssm->fssm_Unit;
	flags = fssm->fssm_Flags;
	device = 1 + BADDR(fssm->fssm_Device);

	if ((diskchgintbit = AllocSignal(-1)) >= 0)
	{
		*diskchgsig_bit = diskchgintbit;
		if ((glob->diskport = CreateMsgPort()))
		{
			if ((glob->diskioreq = CreateIORequest(glob->diskport, sizeof(struct IOExtTD))))
			{
				if (OpenDevice(device, unit, (struct IORequest *)glob->diskioreq, flags) == 0)
				{
					kprintf("\tDevice successfully opened\n");
					if ((glob->diskchgreq = AllocVec(sizeof(struct IOExtTD), MEMF_PUBLIC)))
					{
						CopyMem(glob->diskioreq, glob->diskchgreq, sizeof(struct IOExtTD));
						if ((glob->diskchgintdata = AllocVec(sizeof(struct IntData), MEMF_PUBLIC)))
						{
							struct IntData *DiskChangeIntData = glob->diskchgintdata;

							/* fill interrupt data */
							DiskChangeIntData->glob = glob;
							DiskChangeIntData->task = glob->ourtask;
							DiskChangeIntData->signal = 1 << diskchgintbit;
#ifdef __MORPHOS__
							DiskChangeIntData->InterruptFunc.Trap = TRAP_LIB;
							DiskChangeIntData->InterruptFunc.Extension = 0;
							DiskChangeIntData->InterruptFunc.Func =(void (*)(void))DiskChangeIntHandler;
#endif
							DiskChangeIntData->Interrupt.is_Node.ln_Type = NT_INTERRUPT;
							DiskChangeIntData->Interrupt.is_Node.ln_Pri = 0;
							DiskChangeIntData->Interrupt.is_Node.ln_Name = "MACFS";
							DiskChangeIntData->Interrupt.is_Data = DiskChangeIntData;
#ifdef __MORPHOS__
							DiskChangeIntData->Interrupt.is_Code = (void (*)(void))&DiskChangeIntData->InterruptFunc;
#else
							DiskChangeIntData->Interrupt.is_Code = (void (*)(void))DiskChangeIntHandler;
#endif

							/* fill io request data */
							glob->diskchgreq->iotd_Req.io_Command = TD_ADDCHANGEINT;
							glob->diskchgreq->iotd_Req.io_Data = &DiskChangeIntData->Interrupt;
							glob->diskchgreq->iotd_Req.io_Length = sizeof(struct Interrupt);
							glob->diskchgreq->iotd_Req.io_Flags = 0;
							SendIO((struct IORequest*)glob->diskchgreq);
							kprintf("\tDisk change interrupt handler installed\n");
							return 0;
						}
						else
							err = ERROR_NO_FREE_STORE;
							FreeVec(glob->diskchgreq); 	
					}
					else
						err = ERROR_NO_FREE_STORE;
					CloseDevice((struct IORequest *)glob->diskioreq);
				}
				else
					err = ERROR_DEVICE_NOT_MOUNTED;
				DeleteIORequest(glob->diskioreq);
				glob->diskioreq = NULL;
			}
			else
				err = ERROR_NO_FREE_STORE;
			DeleteMsgPort(glob->diskport);
			glob->diskport = NULL;
		}
		else
			err = ERROR_NO_FREE_STORE;
		FreeSignal(diskchgintbit);
		*diskchgsig_bit = 0;
	}
	else
		err = ERROR_NO_FREE_STORE;
	return err;
}

void CleanupDiskHandler(ULONG diskchgsig_bit)
{
	kprintf("\tFreeing handler resources:\n");

	/* remove disk change interrupt */
	glob->diskchgreq->iotd_Req.io_Command = TD_REMCHANGEINT;
	glob->diskchgreq->iotd_Req.io_Data = glob->diskchgintdata;
	glob->diskchgreq->iotd_Req.io_Length = sizeof(struct Interrupt);
	glob->diskchgreq->iotd_Req.io_Flags = 0;
	DoIO((struct IORequest*)glob->diskchgreq);
	kprintf("\tDisk change interrupt handler removed\n");

	CloseDevice((struct IORequest *)glob->diskioreq);
	DeleteIORequest(glob->diskioreq);
	FreeVec(glob->diskchgreq);
	FreeVec(glob->diskchgintdata);
	DeleteMsgPort(glob->diskport);
	kprintf("\tDevice closed\n");

	glob->diskioreq = NULL;
	glob->diskchgreq = NULL;
	glob->diskchgintdata = NULL;
	glob->diskport = NULL;

	FreeSignal(diskchgsig_bit);
	kprintf("\tDone.\n");
}

